"""Logger config."""

import sys

from loguru import logger


def configure_logger() -> None:
    """Logger config."""

    # Configure `stderr` logger; not reconfigurable.

    logger.remove()
    logger.add(
        sys.stderr,
        level="TRACE",
        format="|".join(
            [
                "{time:HH:mm:ss.SSS}",
                "{thread.name}:{name}.{function}:{line}",
                "{level}",
                "{message}",
            ]
        ),
    )

    add_logging_levels()


def add_logging_levels() -> None:
    """Add custom logging levels."""

    # pylint: disable=too-many-statements

    # remove bold from loguru default colors
    for lvl in logger._core.levels.values():  # noqa: protected-access
        logger.level(lvl.name, color=lvl.color.replace("<bold>", ""))

    # set severity of custom levels relative to the builtins
    _always = _warn = logger.level("WARNING").no
    _info = logger.level("INFO").no
    _debug = logger.level("DEBUG").no
    _trace = logger.level("TRACE").no

    # black blue cyan green magenta red white yellow
    # bold dim normal hide italic blink strike underline reverse
    # logger.level('DEBUG', no=_debug, color='<cyan>')
    # logger.level('TRACE', color='<white>')

    # admin
    logger.level("ADMIN", no=_always, color="<magenta><bold><italic>")
    logger.level("KICK", no=_always, color="<red><italic><reverse>")

    # users
    logger.level("ADDUSER", no=_always, color="<yellow><italic>")
    logger.level("Inactive", no=_always, color="<yellow>")

    # lobby
    logger.level("ADDLOBBY", no=_trace, color="<yellow><italic><bold>")

    # queues
    logger.level("CLEAR", no=_info, color="<magenta>")
    logger.level("EMPTY", no=_info, color="<magenta><reverse>")
    logger.level("PUSH", no=_info, color="<magenta><italic>")
    logger.level("PUSHLEFT", no=_info, color="<magenta><italic><bold>")
    logger.level("POP", no=_info, color="<magenta><italic>")
    logger.level("POPLEFT", no=_info, color="<magenta><italic><bold>")

    logger.level("SPAMS", no=_always, color="<white>")
    logger.level("KICKS", no=_always, color="<white><italic>")

    # gameplay
    logger.level("CAPBLU", no=_always, color="<cyan>")
    logger.level("CAPRED", no=_always, color="<red>")
    logger.level("CHAT", no=_always, color="<green>")
    logger.level("CHATBLU", no=_always, color="<cyan>")
    logger.level("CHATRED", no=_always, color="<red>")
    logger.level("CONNECT", no=_always, color="<magenta>")
    logger.level("DEFBLU", no=_always, color="<cyan>")
    logger.level("DEFRED", no=_always, color="<red>")
    logger.level("KILL", no=_always, color="<green>")
    logger.level("KILLBLU", no=_always, color="<cyan>")
    logger.level("KILLRED", no=_always, color="<red>")
    logger.level("STATUS", no=_trace, color="<magenta>")
    logger.level("TEAMCHAT", no=_always, color="<green><italic>")
    logger.level("TEAMCHATBLU", no=_always, color="<cyan><italic>")
    logger.level("TEAMCHATRED", no=_always, color="<red><italic>")
    logger.level("PERK-ON", no=_always, color="<green><italic>")
    logger.level("PERK-OFF", no=_always, color="<green>")

    # logger.level('action', no=_debug,  color='<blue><bold>')
    # logger.level('assign', no=_debug,  color='<green>')
    logger.level("div", no=_debug, color="<cyan>")
    logger.level("duel", no=_info, color="<magenta>")
    logger.level("exclude", no=_trace, color="<white>")
    logger.level("help", no=_always, color="<magenta>")
    logger.level("ignore", no=_debug, color="<white>")
    logger.level("injected", no=_always, color="<yellow><reverse><italic>")
    logger.level("logline", no=_debug, color="<yellow>")
    logger.level("nextline", no=_warn, color="<yellow>")
    logger.level("console", no=_always, color="<green><bold>")
    logger.level("hacker", no=_always, color="<yellow><reverse>")
    logger.level("regex", no=_trace, color="<magenta>")
    logger.level("report", no=_always, color="<cyan>")
    logger.level("server", no=_trace, color="<cyan>")
    logger.level("QVALVE", no=_always, color="<cyan><italic>")
    logger.level("toggle", no=_debug, color="<green><bold><italic>")
    logger.level("user", no=_debug, color="<green>")
    logger.level("RED", no=_debug, color="<red>")
    logger.level("BLU", no=_debug, color="<cyan>")

    logger.level("MILENKO", no=_always, color="<yellow><bold><italic><reverse>")
    logger.level("CHEATER", no=_always, color="<yellow><bold><italic>")
    logger.level("SUSPECT", no=_always, color="<magenta><bold><italic>")
    logger.level("RACIST", no=_always, color="<red><bold><italic><reverse>")
    logger.level("FUZZ", no=_always, color="<yellow><reverse>")
    logger.level("ALIAS", no=_always, color="<cyan><reverse>")
    logger.level("NOTIFY", no=_always, color="<yellow><reverse>")



"""Team Fortress 2 Console Monitor."""

import csv
import curses
import re
import threading
from pathlib import Path
from pprint import pformat

import libcurses
from loguru import logger

from tf2mon.admin import Admin
from tf2mon.command import Command, CommandManager
from tf2mon.conlog import Conlog
from tf2mon.gameplay import Gameplay
from tf2mon.hacker import HackerAttr, HackerManager
from tf2mon.msgqueue import MsgQueueManager
from tf2mon.role import Role
from tf2mon.spammer import Spammer
from tf2mon.steamweb import SteamWebAPI
from tf2mon.ui import UI
from tf2mon.user import Team
from tf2mon.usermanager import UserManager


class Monitor:
    """Team Fortress 2 Console Monitor."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, cli) -> None:
        """Initialize monitor."""

        self.options = cli.options
        self.config = cli.config

        #
        self.steam_web_api = SteamWebAPI(
            dbpath=self.options.players,
            webapi_key=self.config.get("webapi_key"),
        )

        # Location of TF2 `exec` scripts.
        self.tf2_scripts_dir = Path(self.options.tf2_install_dir, "cfg", "user")
        if not self.tf2_scripts_dir.is_dir():
            logger.warning(f"Missing TF2 scripts dir `{self.tf2_scripts_dir}`")

        # Monitor aliases and key bindings.
        self.path_static_script = self.tf2_scripts_dir / "tf2mon.cfg"  # created once

        # "Send" to TF2 through `msgqueues`.
        self.path_dynamic_script = self.tf2_scripts_dir / "tf2mon-pull.cfg"  # created often
        self.msgqueues = MsgQueueManager(self, self.path_dynamic_script)

        # "Receive" from TF2 through `conlog`.
        # Wait for con_logfile to exist, then open it.
        self.conlog = Conlog(
            self.options.con_logfile,
            exclude_file=self.options.exclude_file,
            rewind=self.options.rewind,
            follow=self.options.follow,
            inject_cmds=self.options.inject_cmds,
            inject_file=self.options.inject_file,
        )

        # this application's admin console
        self.admin = Admin(self)

        #
        self.hackers = HackerManager(
            base=self.options.hackers_base,
            local=self.options.hackers_local,
        )

        #
        self.roles = Role.get_roles_by_name()
        self.sniper_role = self.roles["sniper"]
        self.unknown_role = self.roles["unknown"]

        #
        path = Path(__file__).parent / "data" / "weapons.csv"
        logger.info(f"Reading `{path}`")
        with open(path, encoding="utf-8") as _f:
            self.role_by_weapon = {weapon: self.roles[role] for role, weapon in csv.reader(_f)}

        #
        path = Path(__file__).parent / "data" / "racist.txt"
        logger.info(f"Reading `{path}`")
        lines = path.read_text(encoding="utf-8").splitlines()
        self._racist_re = (
            re.compile("|".join(lines), flags=re.IGNORECASE) if len(lines) > 0 else None
        )

        # users
        self.users = None

        # chats
        self.chats = []

        # message queues
        self.kicks = self.msgqueues.addq("kicks")
        self.spams = self.msgqueues.addq("spams")

        #
        self.spammer = Spammer(self)

        #
        self.commands: CommandManager = self._commands()

        # admin command handlers
        self.regex_list = self.admin.regex_list

        # gameplay handlers
        self.gameplay = Gameplay(self)
        self.regex_list += self.gameplay.regex_list

        # function key handlers
        self.regex_list += self.commands.get_regex_list()

        #
        self.me = self.my = None
        self.ui = None
        self.reset_game()

    def run(self):
        """Run the Monitor."""
        libcurses.wrapper(self._run)

    def _run(self, win):
        self.commands.register_curses_handlers()
        # Build user-interface
        self.ui = UI(self, win)
        self.reset_game()

        # no need for threads if exiting at end of conlog
        if not self.options.follow:
            self.gameplay.repl()
            return

        # Read from conlog, write to display.
        thread = threading.Thread(name="GAME", target=self.gameplay.repl, daemon=True)
        thread.start()

        # main thread reads from keyboard/mouse, and writes to display
        # threading.current_thread().name = "MAIN"
        self.admin.repl()

    def reset_game(self):
        """Start new game."""

        logger.success("RESET GAME")

        self.users = UserManager(self)
        self.me = self.my = self.users.find_username(self.config["player_name"])
        self.me.assign_team(Team.BLU)
        self.my.display_level = "user"
        self.chats = []
        self.msgqueues.clear()

    def breakpoint(self):
        """Drop into python debugger."""

        if self.conlog.is_eof:  # don't do this when replaying logfile from start
            curses.reset_shell_mode()
            breakpoint()  # pylint: disable=forgotten-debug-statement
            curses.reset_prog_mode()

    def kick_my_last_killer(self, attr):
        """Kick the last user who killed the operator."""

        if self.my.last_killer:
            self.my.last_killer.kick(attr)
        else:
            logger.warning("no last killer")

    def is_racist_text(self, text):
        """Return True if this user appears to be racist."""
        return self._racist_re.search(text)

    def dump(self) -> None:
        """Dump stuff."""

        # logger.success(pformat(self.__dict__))
        logger.success(pformat(self.users.__dict__))
        # pylint: disable=protected-access
        for user in self.users._users_by_username.values():
            logger.success(pformat(user.__dict__))

    @property
    def toggling_enabled(self) -> bool:
        """Return True if toggling is enabled.

        Don't allow toggling when replaying a game (`--rewind`),
        unless `--toggles` is also given... or if single-stepping

        This is checked by keys that alter the behavior of gameplay;
        it is not checked by keys that only alter the display.
        """

        return self.conlog.is_eof or self.options.toggles or self.admin.is_single_stepping

    def _commands(self) -> CommandManager:
        """Init and return `CommandManager`."""

        commands = CommandManager()
        commands.bind(self._cmd_help, "F1")
        commands.bind(self._cmd_motd, "Ctrl+F1")
        commands.bind(self._cmd_debug_flag, "F2")
        commands.bind(self._cmd_taunt_flag, "F3")
        commands.bind(self._cmd_show_kd, "F4")
        commands.bind(self._cmd_user_panel, "F5")
        commands.bind(self._cmd_join_other_team, "F6")
        commands.bind(self._cmd_sort_order, "F7")
        commands.bind(self._cmd_log_location, "F8")
        commands.bind(self._cmd_log_level, "Shift+F8")
        commands.bind(self._cmd_reset_padding, "Ctrl+F8")
        commands.bind(self._cmd_grid_layout, "F9")
        commands.bind(self._cmd_show_debug, "KP_INS")
        commands.bind(self._cmd_single_step, "KP_DEL")
        commands.bind(self._cmd_kick_last_cheater, "[", game_only=True)
        commands.bind(self._cmd_kick_last_racist, "]", game_only=True)
        commands.bind(self._cmd_kick_last_suspect, "\\", game_only=True)

        # numpad
        commands.bind(self._cmd_kicks_pop, "KP_HOME")
        commands.bind(self._cmd_kicks_clear, "KP_LEFTARROW")
        commands.bind(self._cmd_kicks_popleft, "KP_END")
        commands.bind(self._cmd_pull, "KP_UPARROW")
        commands.bind(self._cmd_clear_queues, "KP_5")
        commands.bind(self._cmd_push, "KP_DOWNARROW")
        commands.bind(self._cmd_spams_pop, "KP_PGUP")
        commands.bind(self._cmd_spams_clear, "KP_RIGHTARROW")
        commands.bind(self._cmd_spams_popleft, "KP_PGDN")

        if self.tf2_scripts_dir.is_dir():
            logger.info(f"Writing `{self.path_static_script}`")
            script = commands.as_tf2_exec_script(
                str(self.path_static_script.relative_to(self.tf2_scripts_dir.parent)),
                str(self.path_dynamic_script.relative_to(self.tf2_scripts_dir.parent)),
            )
            self.path_static_script.write_text(script, encoding="utf-8")
        else:
            logger.warning(f"Not writing `{self.path_static_script}`")

        return commands

    @staticmethod
    def _on_off(key, value):
        return key.upper() if value else key

    def _cmd_help(self) -> Command:

        return Command(
            name="HELP",
            status=lambda: "HELP",
            handler=lambda m: self.ui.show_help(),
        )

    def _cmd_motd(self) -> Command:

        return Command(
            name="MOTD",
            handler=lambda m: self.ui.show_motd(),
        )

    def _cmd_debug_flag(self) -> Command:
        def _action() -> None:
            if self.toggling_enabled:
                _ = self.ui.debug_flag.toggle
                self.ui.show_status()

        return Command(
            name="TOGGLE-DEBUG",
            status=lambda: self._on_off("debug", self.ui.debug_flag.value),
            handler=lambda m: _action(),
        )

    def _cmd_taunt_flag(self) -> Command:
        def _action() -> None:
            if self.toggling_enabled:
                _ = self.ui.taunt_flag.toggle
                self.ui.show_status()

        return Command(
            name="TOGGLE-TAUNT",
            status=lambda: self._on_off("taunt", self.ui.taunt_flag.value),
            handler=lambda m: _action(),
        )

    def _cmd_show_kd(self) -> Command:
        def _action() -> None:
            if self.toggling_enabled:
                _ = self.ui.show_kd.toggle
                self.ui.show_status()

        return Command(
            name="TOGGLE-KD",
            status=lambda: self._on_off("kd", self.ui.show_kd.value),
            handler=lambda m: _action(),
        )

    def _cmd_user_panel(self) -> Command:
        def _action() -> None:
            _ = self.ui.user_panel.toggle
            self.ui.update_display()

        return Command(
            name="TOGGLE-USER-PANEL",
            status=lambda: self.ui.user_panel.value.name,
            handler=lambda m: _action(),
        )

    def _cmd_join_other_team(self) -> Command:
        def _action() -> None:
            if self.toggling_enabled:
                self.me.assign_team(self.my.opposing_team)
                self.ui.update_display()

        return Command(
            name="SWITCH-MY-TEAM",
            status=lambda: self.my.team.name if self.my.team else "blu",
            handler=lambda m: _action(),
        )

    def _cmd_sort_order(self) -> Command:

        return Command(
            name="TOGGLE-SORT",
            status=lambda: self.ui.sort_order.value.name,
            handler=lambda m: (
                self.ui.set_sort_order(self.ui.sort_order.toggle),
                self.ui.update_display(),
            ),
        )

    def _cmd_log_level(self) -> Command:

        return Command(
            name="TOGGLE-LOG-LEVEL",
            status=lambda: self.ui.log_level.value.name,
            handler=lambda m: (
                self.ui.cycle_log_level(),
                self.ui.show_status(),
            ),
        )

    def _cmd_log_location(self) -> Command:

        return Command(
            name="TOGGLE-LOG-LOCATION",
            status=lambda: self.ui.log_location.value.name,
            handler=lambda m: (
                self.ui.cycle_log_location(),
                self.ui.show_status(),
            ),
        )

    def _cmd_reset_padding(self) -> Command:

        return Command(
            name="RESET-PADDING",
            handler=lambda m: (
                self.ui.logsink.reset_padding(),
                logger.info("padding reset"),
            ),
        )

    def _cmd_grid_layout(self) -> Command:

        return Command(
            name="TOGGLE-LAYOUT",
            status=lambda: self.ui.grid_layout.value.name,
            handler=lambda m: self.ui.cycle_grid_layout(),
        )

    def _cmd_single_step(self) -> Command:

        return Command(
            name="SINGLE-STEP",
            handler=lambda m: self.admin.start_single_stepping(),
        )

    def _cmd_show_debug(self) -> Command:

        return Command(
            name="SHOW-DEBUG",
            handler=lambda m: self.ui.show_debug(),
        )

    def _cmd_kick_last_cheater(self) -> Command:

        return Command(
            name="KICK-LAST-CHEATER",
            status=lambda: HackerAttr.CHEATER.name,
            handler=lambda m: self.kick_my_last_killer(HackerAttr.CHEATER),
        )

    def _cmd_kick_last_racist(self) -> Command:

        return Command(
            name="KICK-LAST-RACIST",
            status=lambda: HackerAttr.RACIST.name,
            handler=lambda m: self.kick_my_last_killer(HackerAttr.RACIST),
        )

    def _cmd_kick_last_suspect(self) -> Command:

        return Command(
            name="KICK-LAST-SUSPECT",
            status=lambda: HackerAttr.SUSPECT.name,
            handler=lambda m: self.kick_my_last_killer(HackerAttr.SUSPECT),
        )

    # --------------------------------------------------------------------------
    # Numpad
    #                      kicks     admin     spams
    #                        |         |         |
    #                        v         v         v
    #                   +-----------------------------+
    #        last/      |         |         |         |
    #        latest --> |   pop   |  pull   |   pop   |
    #                   |         |         |         |
    #                   |---------+---------+---------|
    #                   |         |         |         |
    #                   |  clear  |  clear  |  clear  |
    #                   |         |  both   |         |
    #                   |---------+---------+---------|
    #        first/     |         |         |         |
    #        oldest --> | popleft |  push   | popleft |
    #                   |         |         |         |
    #                   +-----------------------------+

    def _cmd_kicks_pop(self) -> Command:

        return Command(
            name="KICKS-POP",
            handler=lambda m: (
                self.kicks.pop(),
                self.ui.refresh_kicks(),
            ),
            action="tf2mon_kicks_pop",
        )

    def _cmd_kicks_clear(self) -> Command:

        return Command(
            name="KICKS-CLEAR",
            handler=lambda m: (
                self.kicks.clear(),
                self.ui.refresh_kicks(),
            ),
            action="tf2mon_kicks_clear",
        )

    def _cmd_kicks_popleft(self) -> Command:

        return Command(
            name="KICKS-POPLEFT",
            handler=lambda m: (
                self.kicks.popleft(),
                self.ui.refresh_kicks(),
            ),
            action="tf2mon_kicks_popleft",
        )

    def _cmd_pull(self) -> Command:

        return Command(
            name="PULL",
            # handler=lambda m: logger.trace('pull'),
            action="tf2mon_pull",
        )

    def _cmd_clear_queues(self) -> Command:

        return Command(
            name="CLEAR-QUEUES",
            handler=lambda m: (
                self.kicks.clear(),
                self.ui.refresh_kicks(),
                self.spams.clear(),
                self.ui.refresh_spams(),
            ),
            action="tf2mon_clear_queues",
        )

    def _cmd_push(self) -> Command:

        return Command(
            name="PUSH",
            # handler=lambda m: logger.trace('push'),
            action="tf2mon_push",
        )

    def _cmd_spams_pop(self) -> Command:

        return Command(
            name="SPAMS-POP",
            handler=lambda m: (
                self.spams.pop(),
                self.ui.refresh_spams(),
            ),
            action="tf2mon_spams_pop",
        )

    def _cmd_spams_clear(self) -> Command:

        return Command(
            name="SPAMS-CLEAR",
            handler=lambda m: (
                self.spams.clear(),
                self.ui.refresh_spams(),
            ),
            action="tf2mon_spams_clear",
        )

    def _cmd_spams_popleft(self) -> Command:

        return Command(
            name="SPAMS-POPLEFT",
            handler=lambda m: (
                self.spams.popleft(),
                self.ui.refresh_spams(),
            ),
            action="tf2mon_spams_popleft",
        )



"""Extend `list` with a `toggle` method to cycle through the values with each access."""


class Cycle(list):
    """Circular list to return next item with each access.

    Extend `list` with a `toggle` method to cycle through the values with
    each access.
    """

    def __init__(self, name, values):
        """Create new `Cycle`.

        Args:
            name:   name of instance.
            values: list of values to cycle through.
        """

        super().__init__()

        self.name = name
        assert values
        self._values = list(values)
        _len = len(self._values)
        assert _len
        self._max = _len - 1
        self._idx = 0

    def __setitem__(self, index, value):
        # ignore index
        if value not in self._values:
            raise KeyError(f"value {value!r} not in {self.name!r}")
        while self._values[self._idx] != value:
            self.next  # pylint: disable=pointless-statement
        return value

    def start(self, value):
        """Set starting `value`."""
        return self.__setitem__(None, value)  # noqa

    def __getitem__(self, index):
        return self._values[index]

    def __repr__(self):
        return str(self._values)

    @property
    def prev(self):
        """Return previous value."""
        if self._max:
            self._idx = self._max if self._idx == 0 else self._idx - 1
        return self._values[self._idx]

    @property
    def next(self):
        """Return next value."""
        if self._max:
            self._idx = 0 if self._idx == self._max else self._idx + 1
        return self._values[self._idx]

    cycle = next
    toggle = next

    @property
    def value(self):
        """Return current value."""
        return self._values[self._idx]


#    @value.setter
#    def value(self, value):
#        if value not in self._values:
#            raise KeyError(f'bad value {value!r}')
#        while self._values[self._idx] != value:
#            self.next

Toggle = Cycle



"""Function Keys."""

import curses
import re
from typing import ClassVar

# Values for `modifier`; also used as `label` prefix.
MOD_BASE = ""
MOD_CTRL = "c"
MOD_SHIFT = "s"


class FKey:
    """The simultaneous pressing of a base key and an optional modifier key."""

    pattern: ClassVar[re.Pattern] = re.compile(r"^(?:(SHIFT|CTRL)\+)?(F)?(.+)$")

    curses_from_game: ClassVar[dict[str, int]] = {
        "KP_HOME": curses.KEY_HOME,
        "KP_END": curses.KEY_END,
        "KP_UPARROW": curses.KEY_UP,
        "KP_DOWNARROW": curses.KEY_DOWN,
        "KP_LEFTARROW": curses.KEY_LEFT,
        "KP_RIGHTARROW": curses.KEY_RIGHT,
        "KP_PGUP": curses.KEY_PPAGE,
        "KP_PGDN": curses.KEY_NPAGE,
        "KP_5": curses.KEY_B2,
        "KP_INS": curses.KEY_IC,
        "KP_DEL": curses.KEY_DC,
        # "xxx": curses.KEY_SR,
        # "xxx": curses.KEY_SF,
        # "xxx": curses.KEY_SLEFT,
        # "xxx": curses.KEY_SRIGHT,
        # "xxx": curses.KEY_SDC,  # can't do KEY_SIC
    }

    def __init__(self, keyspec: str):
        """Init `FKey` from `keyspec`.

        Keyname in TF2 terms, with optional modifier "shift+" or "ctrl+" (not both).
        e.g., "A", "shift+A", "ctrl+KP_LEFTARROW".

        Case-insensitive:
        >>> FKey("a").__dict__ == FKey("A").__dict__
        True
        >>> FKey("shift+a").__dict__ == FKey("shift+A").__dict__
        True

        Note that:
        >>> FKey("shift+a").__dict__ != FKey("A").__dict__
        True
        >>> FKey("shift+A").__dict__ != FKey("A").__dict__
        True
        """

        # pylint: disable=too-many-branches

        if not keyspec:
            raise ValueError("keyspec", keyspec)
        self.keyspec = keyspec.upper()

        matched = self.pattern.match(self.keyspec)
        if not matched:
            raise ValueError("keyspec", self.keyspec)

        mod, eff, key = matched.groups()
        self.name: str = (eff or "") + key  # TF2 terms: "B", "KP_LEFTARROW", "F1"

        self.modifier: str = MOD_BASE
        if mod:
            if mod == "SHIFT":
                self.modifier = MOD_SHIFT
            elif mod == "CTRL":
                self.modifier = MOD_CTRL
            else:
                raise ValueError("keyspec", self.keyspec)

        self.key: int = None  # ord("B"), curses.KEY_LEFT, curses.KEY_F1
        if eff:
            self.key = curses.KEY_F0
            try:
                self.key += int(key)
            except ValueError as exc:
                raise ValueError("keyspec", self.keyspec) from exc
        elif len(key) == 1:
            self.key = ord(key)
        elif (key := self.curses_from_game.get(key)) is not None:
            self.key = key
        else:
            raise ValueError("keyspec", self.keyspec)

        if eff:
            if self.modifier == MOD_SHIFT:
                self.key += 12
            elif self.modifier == MOD_CTRL:
                self.key += 24

        self.label: str = self.modifier + self.name  # display name

    def __repr__(self):
        return str(self.__dict__)

    @property
    def is_base(self) -> bool:
        """Return True if only `base` key was pressed."""
        return self.modifier == MOD_BASE

    @property
    def is_ctrl(self) -> bool:
        """Return True if `ctrl` was also pressed."""
        return self.modifier == MOD_CTRL

    @property
    def is_shift(self) -> bool:
        """Return True if `shift` was also pressed."""
        return self.modifier == MOD_SHIFT



"""Role (job, function, position) of `User`."""

from dataclasses import dataclass


@dataclass
class Role:
    """Role (job, function, position) of `User`."""

    name: str  # match tokens in `con_logfile` and `weapons.csv`
    display: str  # for scoreboard

    @staticmethod
    def get_roles_by_name() -> dict:
        """Return `dict` of `Role`s indexed by name."""

        roles = {}
        for display, name in [
            ("C", "scout"),
            ("S", "soldier"),
            ("P", "pyro"),
            ("D", "demo"),
            ("E", "engineer"),
            ("H", "heavy"),
            ("M", "medic"),
            ("N", "sniper"),
            ("Y", "spy"),
            ("?", "unknown"),
        ]:
            roles[name] = Role(name, display)

        return roles

    def get_weapon_state(self, weapon: str, crit: bool, perk: str) -> str:
        """Return hashable key for given weapon_state.

        Also used as the formatted display value by scoreboard.
        """

        crit = "+crit" if crit else ""
        perk = "+" + perk if perk else ""

        return " ".join(
            [
                f"{self.name:8}",  # 8=len("engineer")
                f"{crit:5}",  # 5=len("+crit")
                f"{weapon:26}",  # 26=len("tf_projectile_pipe_remote")
                perk,  # 20=len("Infinite Double Jump")
            ]
        ).rstrip()



"""Communication mechanism to "send" data to the game."""

from collections import deque

from loguru import logger

import tf2mon


class MsgQueue:
    """Communication mechanism to "send" data to the game."""

    def __init__(self, monitor, name):
        """Create new `MsgQueue`.

        Args:
            monitor: the monitor.
            name:   name of message queue.
        """

        self.monitor = monitor
        self.name = name
        self.msgs = deque()

    def push(self, msg):
        """Append message to end of queue."""

        self.msgs.append(msg)
        logger.opt(depth=1).log("PUSH", msg)

    def pushleft(self, msg):
        """Append message to other end of queue."""

        self.msgs.insert(0, msg)
        logger.opt(depth=1).log("PUSHLEFT", msg)

    def pop(self):
        """Remove and return message from end of queue."""

        self._pop("POP", self.msgs.pop() if self.msgs else None)

    def popleft(self):
        """Remove and return message from other end of queue."""

        self._pop("POPLEFT", self.msgs.popleft() if self.msgs else None)

    def _pop(self, action, msg):

        if not self.msgs:
            logger.log("EMPTY", f" {self.name} ".center(80, "-"))
        else:
            logger.opt(depth=1).log(action, msg)

    def clear(self):
        """Remove all messages from the queue."""

        self.msgs.clear()
        logger.opt(depth=1).log("CLEAR", self.name)

    def aliases(self):
        """Return message queue function key aliases.

        Create and return definitions of the `_tf2mon` prefixed-
        aliases that are called from `~/tf2/cfg/tf2-monitor.cfg`.

                                                      --- these aliases ---
        alias tf2mon_kicks_pop          "tf2mon_pull; _tf2mon_kicks_pop"
        alias tf2mon_kicks_popleft      "tf2mon_pull; _tf2mon_kicks_popleft"
        alias tf2mon_spams_pop          "tf2mon_pull; _tf2mon_spams_pop"
        alias tf2mon_spams_popleft      "tf2mon_pull; _tf2mon_spams_popleft"
        """

        # Create acknowledgement commands like this:
        #
        #   echo TF2MON-KICKS-POP
        #   echo TF2MON-KICKS-POPLEFT
        #   echo TF2MON-SPAMS-POP
        #   echo TF2MON-SPAMS-POPLEFT

        if self.msgs:
            last, first = self.msgs[-1], self.msgs[0]
            last_ack = f" ; echo {tf2mon.APPTAG}{self.name.upper()}-POP"
            first_ack = last_ack + "LEFT"
        else:
            _echo = "say" if self.monitor.ui.debug_flag.value else "echo"
            last = first = f"{_echo} the {self.name} queue is empty."
            last_ack = first_ack = ""

        return [
            f'alias _tf2mon_{self.name}_pop "{last}{last_ack}"',
            f'alias _tf2mon_{self.name}_popleft "{first}{first_ack}"',
        ]


class MsgQueueManager:
    """Collection of `MsgQueue`s."""

    def __init__(self, monitor, path):
        """Initialize collection of `MsgQueue`s."""

        self.monitor = monitor
        # list(MsgQueue)
        self._queues = []
        self._file = None
        if path and path.parent.is_dir():
            # pylint: disable=consider-using-with
            self._file = open(path, "w", encoding="utf-8")

    def addq(self, name):
        """Create `MsgQueue` with `name`, add to collection and return it."""

        queue = MsgQueue(self.monitor, name)
        self._queues.append(queue)
        return queue

    def clear(self):
        """Clear all message queues."""

        for queue in self._queues:
            queue.clear()

    def send(self):
        """Send data to tf2 by writing aliases to an `exec` script."""

        if not self._file:
            return

        self._file.seek(0)
        self._file.truncate()
        for queue in self._queues:
            print("\n".join(queue.aliases()), file=self._file)
        self._file.flush()



"""Wrapper for `ISteamUser.GetPlayerSummaries`.

See https://developer.valvesoftware.com/wiki/Steam_Web_API
"""
import time

import humanize
import steam.steamid
from loguru import logger


class SteamPlayer:
    """Represent an ['response']['players'] element from `ISteamUser.GetPlayerSummaries`."""

    # pylint: disable=too-many-instance-attributes

    BOT_NAME = "BOT"
    BOT_S_STEAMID = "1"
    BOT_STEAMID = int(BOT_S_STEAMID)

    def __init__(self, jdoc):
        """Create `SteamPlayer` from json document.

        Args:
            jdoc:   `['response']['players']` element returned by
                    `ISteamUser.GetPlayerSummaries`.
        """

        self.steamid = steam.steamid.SteamID(jdoc.get("steamid"))
        self.personaname = jdoc.get("personaname")
        self.profileurl = jdoc.get("profileurl")
        self.personastate = jdoc.get("personastate")
        self.realname = jdoc.get("realname")
        self.timecreated = jdoc.get("timecreated")
        self.loccountrycode = jdoc.get("loccountrycode")
        self.locstatecode = jdoc.get("locstatecode")
        self.loccityid = jdoc.get("loccityid")

        #
        if self.profileurl and self.profileurl == self.steamid.community_url + "/":
            # for asthetics only; to avoid clutter
            self.profileurl = None  # indicate long noisy determinable value

        #
        if not self.timecreated:
            logger.warning(f"no timecreated {str(self.__dict__)}")

        now = int(time.time())
        self.age = humanize.naturaldelta(now - self.timecreated) if self.timecreated else ""
        self.mtime = jdoc.get("mtime", now)

    def __repr__(self):
        return str(self.__dict__)

    @property
    def is_legitimate_game_bot(self):
        """Return True if this is a legitimate game BOT; not a hacker."""

        return self.steamid.id == self.BOT_STEAMID


def steamid_from_str(s_steamid):
    """Return `steam.steamid.SteamID` from string in STATUS and LOBBY lines.

    Return None if invalid.

    Args:
        s_steamid: string, decimal number or "BOT".
    """

    if s_steamid == SteamPlayer.BOT_NAME:
        # create a dummy steamid for this bot; (not a hacker, a real game bot)
        s_steamid = SteamPlayer.BOT_S_STEAMID

    steamid = steam.steamid.SteamID(s_steamid)
    if not steamid.is_valid():
        logger.error(f"invalid steamid {s_steamid}")
        return None

    return steamid



"""Grid layouts."""

from enum import Enum

from tf2mon.layouts.default import DefaultLayout
from tf2mon.layouts.full import FullLayout
from tf2mon.layouts.tall import TallLayout
from tf2mon.layouts.tallchat import TallChatLayout
from tf2mon.layouts.wide import WideLayout

LAYOUT_ENUM = Enum("_layout_enum", "DFLT FULL TALL MRGD WIDE")
LAYOUT_ENUM.__doc__ = "Grid layout."

LAYOUT_CLASSES = {
    LAYOUT_ENUM.DFLT: DefaultLayout,
    LAYOUT_ENUM.FULL: FullLayout,
    LAYOUT_ENUM.TALL: TallLayout,
    LAYOUT_ENUM.MRGD: TallChatLayout,
    LAYOUT_ENUM.WIDE: WideLayout,
}



"""User Interface."""

import contextlib
import curses
import os
import sys
import textwrap
from enum import Enum

import libcurses
from loguru import logger

import tf2mon.layouts
from tf2mon.scoreboard import Scoreboard
from tf2mon.toggle import Toggle
from tf2mon.user import Team, UserState

# from playsound import playsound


# These would have been defined within class UI (because they're only used internally)
# but `pydoc` doesn't display their values when defined there; it does when defined here.

USER_PANEL = Enum("_user_panel_enum", "AUTO DUELS KICKS SPAMS")
USER_PANEL.__doc__ = "Contents of user panel."

LOG_LEVEL = Enum("_lvl_enum", "INFO DEBUG TRACE")
LOG_LEVEL.__doc__ = "Logging level."

LOG_LOCATION = Enum("_loc_enum", "MOD NAM THM THN FILE NUL")
LOG_LOCATION.__doc__ = "Format of logger location field."

SORT_ORDER = Enum("_sort_order_enum", "STEAMID K KD USERNAME")
SORT_ORDER.__doc__ = "Scoreboard sort column."


class UI:
    """User Interface."""

    # pylint: disable=too-many-instance-attributes

    _log_levels = {
        LOG_LEVEL.INFO: "INFO",  # ""
        LOG_LEVEL.DEBUG: "DEBUG",  # "-v"
        LOG_LEVEL.TRACE: "TRACE",  # "-vv"
    }
    log_level = Toggle("_lvl_cycle", LOG_LEVEL)

    _log_locations = {
        LOG_LOCATION.MOD: "{module}.{function}:{line}",
        LOG_LOCATION.NAM: "{name}.{function}:{line}",
        LOG_LOCATION.THM: "{thread.name}:{module}.{function}:{line}",
        LOG_LOCATION.THN: "{thread.name}:{name}.{function}:{line}",
        LOG_LOCATION.FILE: "{file}:{function}:{line}",
        LOG_LOCATION.NUL: None,
    }
    log_location = Toggle("_loc_cycle", LOG_LOCATION)

    def __init__(self, monitor, win: curses.window):
        """Initialize User Interface."""

        self.monitor = monitor
        self.notify_operator = False
        self.sound_alarm = False

        # F2 Toggle Debug (control `say` vs `echo`).
        self.debug_flag = Toggle("_df", [False, True])

        # F3 Enable/disable Taunts and Throes.
        self.taunt_flag = Toggle("_tf", [False, True])
        self.throe_flag = Toggle("_gf", [True, False])

        # F4 Include kd-ratio in messages (`User.moniker`)
        self.show_kd = Toggle("_kd", [False, True])

        # F5 Control User-panel display: Kicks, Spams, Duels and Auto.
        self.user_panel = Toggle("_up", USER_PANEL)

        # options when displaying USER_PANEL.USER
        # self.show_actions = Toggle("_sa", [True, False])
        # if self.show_actions.value:
        #     lines.extend(user.actions)

        # Scoreboard sort column.
        self.sort_order = Toggle("_so", SORT_ORDER)
        self.sort_order.start(SORT_ORDER.KD)

        # create empty grid
        self.grid = libcurses.Grid(win)

        # `self.build_grid` creates these windows, sized to fill `win`.
        self.chatwin_blu: curses.window = None
        self.chatwin_red: curses.window = None
        self.scorewin_blu: curses.window = None
        self.scorewin_red: curses.window = None
        self.user_win: curses.window = None
        self.kicks_win: curses.window = None
        self.spams_win: curses.window = None
        self.duels_win: curses.window = None
        self.logger_win: curses.window = None
        self.status_win: curses.window = None
        self.cmdline_win: curses.window = None

        # the windows may be placed in different arrangements.
        self.grid_layout = Toggle("_grid_layout", tf2mon.layouts.LAYOUT_ENUM)
        self.grid_layout.start(self.monitor.options.layout)

        # `register_builder` 1) calls `build_grid` and 2) configures
        # `KEY_RESIZE` to call it again each time that event occurs.
        self.grid.register_builder(self.build_grid)

        #
        self.logsink = libcurses.Sink(self.logger_win)
        #
        self.logsink.set_verbose(self.monitor.options.verbose)
        self.log_level.start(LOG_LEVEL.__dict__[self.logsink.level])
        #
        self.log_location.start(LOG_LOCATION.MOD)
        self.logsink.set_location(self._log_locations[self.log_location.value])

        self.colormap = libcurses.get_colormap()
        #
        self._scoreboard = Scoreboard(
            self.monitor,
            self.scorewin_blu,
            self.colormap[Team.BLU.name],
            self.scorewin_red,
            self.colormap[Team.RED.name],
        )

        self.set_sort_order(self.sort_order.value)

    def cycle_grid_layout(self):
        """Use next grid layout."""

        _ = self.grid_layout.cycle

        # self.grid.boxes = self.grid.boxes[:1]
        self.grid.handle_term_resized_event()
        self.show_status()

    def build_grid(self):
        """Add boxes to grid.

        Called at init, on KEY_RESIZE events, and when layout changes.
        """

        klass = tf2mon.layouts.LAYOUT_CLASSES[self.grid_layout.value]
        try:
            layout = klass(self.grid)
        except AssertionError:
            curses.endwin()
            msg = "Terminal too small; try `Maximize` and `Ctrl+Minus`."
            logger.error(msg)
            if not os.isatty(sys.stderr.fileno()):
                print(f"ERROR: {msg}")
            sys.exit(1)

        if os.isatty(sys.stderr.fileno()):
            os.close(sys.stderr.fileno())

        if not layout.logger_win:
            raise RuntimeError(f"undefined `logger_win` in {layout}.")
        if not layout.status_win:
            raise RuntimeError(f"undefined `status_win` in {layout}.")
        if not layout.cmdline_win:
            raise RuntimeError(f"undefined `cmdline_win` in {layout}.")

        # promote
        self.chatwin_blu = layout.chatwin_blu
        self.chatwin_red = layout.chatwin_red
        self.scorewin_blu = layout.scorewin_blu
        self.scorewin_red = layout.scorewin_red
        self.user_win = layout.user_win
        self.kicks_win = layout.kicks_win
        self.spams_win = layout.spams_win
        self.duels_win = layout.duels_win
        self.logger_win = layout.logger_win
        self.status_win = layout.status_win
        self.cmdline_win = layout.cmdline_win

        for win in [
            self.chatwin_blu,
            self.chatwin_red,
            self.logger_win,
        ]:
            if win:
                win.scrollok(True)

        for win in [
            self.scorewin_blu,
            self.scorewin_red,
            self.user_win,
            self.kicks_win,
            self.spams_win,
            self.duels_win,
            self.status_win,
        ]:
            if win:
                win.scrollok(False)

        if self.cmdline_win:
            self.cmdline_win.scrollok(True)
            self.cmdline_win.keypad(True)

        self.refresh_chats()
        self.grid.redraw()

    def cycle_log_level(self) -> None:
        """Cycle logging level in logger window."""

        self.logsink.set_level(self._log_levels[self.log_level.cycle])

    def cycle_log_location(self) -> None:
        """Cycle format of location in messages displayed in logger window."""

        self.logsink.set_location(self._log_locations[self.log_location.cycle])

    def set_sort_order(self, sort_order):
        """Set scoreboard sort column."""

        self.monitor.users.set_sort_order(sort_order)
        self._scoreboard.set_sort_order(sort_order)

    def getline(self, prompt=None):
        """Read and return next line from keyboard."""

        self.cmdline_win.erase()
        if prompt:
            self.cmdline_win.addstr(0, 0, prompt)
        return libcurses.getline(self.cmdline_win)

    def update_display(self):
        """Update display."""

        if self.sound_alarm:
            self.sound_alarm = False
            if self.monitor.conlog.is_eof:  # don't do this when replaying logfile from start
                ...
                # playsound('/usr/share/sounds/sound-icons/prompt.wav')
                # playsound('/usr/share/sounds/sound-icons/cembalo-10.wav')
                # curses.flash()
                # curses.beep()

        self.refresh_kicks()
        self.refresh_spams()
        self.refresh_duels(self.monitor.me)
        self.refresh_user(self.monitor.me)
        # chatwin_blu and chatwin_red are rendered from gameplay/_playerchat
        self._scoreboard.show_scores(
            team1=list(self.monitor.users.active_team_users(Team.BLU)),
            team2=list(self.monitor.users.active_team_users(Team.RED)),
        )
        self.show_status()
        self.grid.refresh()

    def refresh_chats(self):
        """Refresh chat panel(s)."""

        for win in [
            self.chatwin_blu,
            self.chatwin_red,
        ]:
            if win:
                win.erase()

        for chat in self.monitor.chats:
            self.show_chat(chat)

    def refresh_kicks(self):
        """Refresh kicks panel."""

        if self.kicks_win:
            # ic(self.kicks_win.getbegyx())
            # ic(self.kicks_win.getmaxyx())
            self._show_lines("KICKS", reversed(self.monitor.kicks.msgs), self.kicks_win)

        # if self.user_win:
        #     self.refresh_user(self.monitor.me)

    def refresh_spams(self):
        """Refresh spams panel."""

        if self.spams_win:
            self._show_lines("SPAMS", reversed(self.monitor.spams.msgs), self.spams_win)

        if self.user_win:
            self.refresh_user(self.monitor.me)

    def refresh_duels(self, user):
        """Refresh duels panel."""

        if self.duels_win:
            self._show_lines("user", self._format_duels(user), self.duels_win)

        if self.user_win:
            self.refresh_user(self.monitor.me)

    def refresh_user(self, user):
        """Refresh user panel."""

        if self.user_win:
            if self.user_panel.value == USER_PANEL.KICKS or (
                self.user_panel.value == USER_PANEL.AUTO and self.monitor.kicks.msgs
            ):
                self._show_lines(
                    "KICKS",
                    reversed(self.monitor.kicks.msgs)
                    if self.monitor.kicks.msgs
                    else ["No Kicks"],
                    self.user_win,
                )
            #
            elif self.user_panel.value == USER_PANEL.SPAMS or (
                self.user_panel.value == USER_PANEL.AUTO and self.monitor.spams.msgs
            ):
                self._show_lines(
                    "SPAMS",
                    reversed(self.monitor.spams.msgs)
                    if self.monitor.spams.msgs
                    else ["No Spams"],
                    self.user_win,
                )
            #
            else:
                self._show_lines("user", self._format_duels(user), self.user_win)

            self.user_win.noutrefresh()

    def user_color(self, user, color):
        """Return `color` to display `user` in scoreboard."""

        if user.display_level:
            color = self.colormap[user.display_level]

        if user == self.monitor.my.last_killer:
            color |= curses.A_BOLD | curses.A_ITALIC

        if user == self.monitor.my.last_victim:
            color |= curses.A_BOLD

        if user.selected:
            color |= curses.A_REVERSE

        if not user.team:
            color |= curses.A_UNDERLINE

        if user.cloner:
            color |= curses.A_ITALIC

        return color

    def show_chat(self, chat):
        """Display (append) `chat` in appropriate team window."""

        win = None
        if chat.user.team == Team.RED:
            win = self.chatwin_red
        if not win:
            win = self.chatwin_blu  # unassigned, or RED in shared window.
        if not win:
            return  # not showing chats

        color = self.colormap[chat.user.team.name if chat.user.team else "user"]
        if chat.teamflag:
            color |= curses.A_UNDERLINE

        if sum(win.getyx()):
            win.addch("\n")
        line = f"{chat.seqno}: {chat.user.username:20.20}: {chat.msg}"
        win.addstr(line, self.user_color(chat.user, color))
        win.noutrefresh()

    def _format_duels(self, user):

        lines = []
        indent = " " * 12  # 12=len("99 and 99 vs")

        lines.append("Duels:")
        for opponent in [x for x in user.opponents.values() if x.state == UserState.ACTIVE]:
            lines.append(f"{user.duel_as_str(opponent, True)} vs {opponent.moniker}")

            if opponent.key in user.nkills_by_opponent_by_weapon:
                for weapon, count in user.nkills_by_opponent_by_weapon[opponent.key].items():
                    lines.append(f"{indent} K {count:2} {weapon}")

            if user.key in opponent.nkills_by_opponent_by_weapon:
                for weapon, count in opponent.nkills_by_opponent_by_weapon[user.key].items():
                    lines.append(f"{indent} D {count:2} {weapon}")

        return lines

    def show_status(self):
        """Update status line."""

        line = self.monitor.commands.get_status_line() + f" UID={self.monitor.my.userid}"

        try:
            self.status_win.addstr(
                0, 0, line, self.colormap["NOTIFY" if self.notify_operator else "STATUS"]
            )
            self.status_win.clrtoeol()
        except curses.error:
            pass
        self.status_win.noutrefresh()

    def _show_lines(self, level, lines, win):

        win.erase()

        with contextlib.suppress(curses.error):
            for line in lines:
                win.addstr(line + "\n", self.colormap[level])
        win.noutrefresh()

    def show_help(self):
        """Show help."""

        for line in (
            textwrap.dedent(
                """
    Press Enter to process next line.
    Enter "b 500" to set breakpoint at line 500.
    Enter "/pattern[/i]" to set search pattern.
    Enter "/" to clear search pattern.
    Enter "c" to continue.
    Enter "quit" or press ^D to quit."
                """
            )
            .strip()
            .splitlines()
        ):
            logger.log("help", line)

    def show_motd(self):
        """Show message of the day."""

        motd = self.monitor.tf2_scripts_dir.parent / "motd.txt"
        logger.log("help", f" {motd} ".center(80, "-"))

        with open(motd, encoding="utf-8") as file:
            for line in file:
                logger.log("help", line.strip())

    def show_debug(self):
        """Debug grid."""

        logger.debug(self.grid)
        for box in self.grid.boxes:
            logger.debug(self.grid.winyx(box))



"""Full layout."""

import libcurses

from tf2mon.baselayout import BaseLayout


class FullLayout(BaseLayout):
    """Full layout.

    +-------------------+------------+---------+
    | score blu         | chat blu   | kicks   |
    |                   |            |         |
    +-------------------+------------+---------+
    | score red         | chat red   | spams   |
    |                   |            |         |
    +--------------+----+------------+---------+
    | user         | logger                    |
    |              |                           |
    +--------------+--------+------------------+
    | status                | cmdline          |
    +-----------------------+------------------+
    """

    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    def __post_init__(self, grid: libcurses.Grid, max_users: int):
        """Build windows."""

        _nlines = int(max_users / 2) + 2 + 1  # 2=borders (top and bottom), 1=header.
        # 124=len(self._scoreboard._formatted_header) + 2=borders (left and right) + 1=padding
        # _ncols = min(124, int(2 * grid.ncols / 3))
        _ncols = int(grid.ncols / 2)

        # top-left
        self.scorewin_blu = grid.box(
            "scorewin_blu",
            nlines=_nlines,
            ncols=_ncols,
            left=grid,
            top=grid,
        )

        # stack scoreboards
        self.scorewin_red = grid.box(
            "scorewin_red",
            nlines=_nlines,
            ncols=_ncols,
            left=grid,
            top2b=self.scorewin_blu,
        )

        # bottom-left
        self.status_win = grid.box(
            "status",
            nlines=3,
            ncols=_ncols,
            left=grid,
            bottom=grid,
        )

        # bottom-right
        self.cmdline_win = grid.box(
            "cmdline",
            nlines=3,
            ncols=0,
            left2r=self.status_win,
            right=grid,
            bottom=grid,
        )

        #
        self.chatwin_blu = grid.box(
            "chatwin_blu",
            nlines=0,
            ncols=60,
            left2r=self.scorewin_blu,
            top=self.scorewin_blu,
            bottom=self.scorewin_blu,
        )

        self.chatwin_red = grid.box(
            "chatwin_red",
            nlines=0,
            ncols=60,
            left2r=self.scorewin_red,
            top=self.scorewin_red,
            bottom=self.scorewin_red,
        )

        #
        self.kicks_win = grid.box(
            "kicks_win",
            nlines=0,
            ncols=0,
            left2r=self.chatwin_blu,
            right=grid,
            top=self.chatwin_blu,
            bottom=self.chatwin_blu,
        )

        self.spams_win = grid.box(
            "spams_win",
            nlines=0,
            ncols=0,
            left2r=self.chatwin_red,
            right=grid,
            top=self.chatwin_red,
            bottom=self.chatwin_red,
        )

        # section 3 - fills gap between sections 2 and 4
        self.user_win = grid.box(
            "user",
            nlines=0,
            ncols=min(62, grid.ncols // 2),  # 62: see get_weapon_state
            left=grid,
            top2b=self.scorewin_red,
            bottom2t=self.status_win,
        )

        self.logger_win = grid.box(
            "logwin",
            nlines=0,
            ncols=0,
            left2r=self.user_win,
            right=grid,
            top=self.user_win,
            bottom=self.user_win,
        )



"""Default layout."""

import libcurses

from tf2mon.baselayout import BaseLayout


class DefaultLayout(BaseLayout):
    """Default layout.

    +-------------------+-----------+----------+
    | score blu         | kicks     | spams    |
    |                   |           |          |
    +-------------------+           |          |
    | score red         |           |          |
    |                   |           |          |
    +-------------------+-----------+----------+
    | chats (merged)    | duels                |
    |                   |                      |
    +-------------------+                      |
    | logger            |                      |
    |                   |                      |
    +-------------------+----------------------+
    | status            | cmdline              |
    +-------------------+----------------------+
    """

    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes

    def __post_init__(self, grid: libcurses.Grid, max_users: int):
        """Build windows."""

        _nlines = int(max_users / 2) + 2 + 1  # 2=borders (top and bottom), 1=header.
        # 124=len(self._scoreboard._formatted_header) + 2=borders (left and right) + 1=padding
        # _ncols = min(124, int(2 * grid.ncols / 3))
        _ncols = int(grid.ncols / 2)

        # top-left
        self.scorewin_blu = grid.box(
            "scorewin_blu",
            nlines=_nlines,
            ncols=_ncols,
            left=grid,
            top=grid,
        )

        # stack scoreboards
        gap_top = self.scorewin_red = grid.box(
            "scorewin_red",
            nlines=_nlines,
            ncols=_ncols,
            left=grid,
            top2b=self.scorewin_blu,
        )

        # -- leave vertical gap --

        # bottom-left
        gap_bottom = self.status_win = grid.box(
            "status",
            nlines=3,
            ncols=_ncols,
            left=grid,
            bottom=grid,
        )

        # -- fill vertical gap with:
        # -- 50% chats
        # -- 50% logger

        _nlines = int(
            ((gap_bottom.getbegyx()[0] - gap_top.getbegyx()[0]) - gap_top.getmaxyx()[0]) / 2
        )

        self.chatwin_blu = grid.box(
            "chatwin_blu",
            nlines=_nlines,
            ncols=_ncols,
            left=grid,
            top2b=self.scorewin_red,
        )

        self.logger_win = grid.box(
            "logwin",
            nlines=0,
            ncols=_ncols,
            left=grid,
            top2b=self.chatwin_blu,
            bottom2t=self.status_win,
        )

        # -------------------------------------------------------------------------------

        # top-right
        gap_top_grid = grid.win

        # -- leave vertical gap --

        # bottom-right
        gap_bottom = self.cmdline_win = grid.box(
            "cmdline",
            nlines=3,
            ncols=0,
            left2r=self.status_win,
            right=grid,
            bottom=grid,
        )

        # -- fill vertical gap with:
        # -- 50% kicks/spams
        # -- 50% duels
        _nlines = int((gap_bottom.getbegyx()[0] - gap_top_grid.getbegyx()[0]) / 2)

        # kicks/spams
        gap_left = self.scorewin_red
        # -- fill horizontal gap with:
        # -- 50% kicks (left)
        # -- 50% spams (right)
        _ncols = int((grid.ncols - gap_left.getmaxyx()[1]) / 2)

        # top-right, left-half
        self.kicks_win = grid.box(
            "kicks_win",
            nlines=_nlines,
            ncols=_ncols,
            left2r=self.chatwin_blu,
            top=grid,
        )

        # top-right, right-half
        self.spams_win = grid.box(
            "spams_win",
            nlines=_nlines,
            ncols=0,
            left2r=self.kicks_win,
            right=grid,
            top=grid,
        )

        # final gap
        self.duels_win = grid.box(
            "duels_win",
            nlines=0,
            ncols=0,
            left2r=self.chatwin_blu,
            right=grid,
            top2b=self.kicks_win,
            bottom2t=self.cmdline_win,
        )



import string

import icecream
import pytest
from icecream import ic

from tf2mon.fkey import FKey

icecream.install()
icecream.ic.configureOutput(prefix="=====>\n", includeContext=True)


@pytest.mark.parametrize(
    ("keyspec"),
    [
        None,
        "",
        "shift",
        "shift+",
        "ctrl",
        "ctrl+",
        "shif+A",
        "ctr+A",
        "ctrl+shift",
        "ctrl+shift+",
        "ctrl+shift+A",
        "+A",
        "+shift+A",
        "+ctrl+A",
    ],
)
def test_fkey_value_error(keyspec):

    with pytest.raises(ValueError, match="keyspec") as err:
        FKey(keyspec)
    assert err.type == ValueError


@pytest.mark.parametrize(
    ("keyspec", "key", "is_shift", "is_ctrl"),
    [(x, ord(x), False, False) for x in string.ascii_uppercase]
    + [(x, ord(x) - (ord("a") - ord("A")), False, False) for x in string.ascii_lowercase]
    + [(f"shift+{x}", ord(x), True, False) for x in string.ascii_uppercase]
    + [
        (f"shift+{x}", ord(x) - (ord("a") - ord("A")), True, False)
        for x in string.ascii_lowercase
    ]
    + [(f"ctrl+{x}", ord(x), False, True) for x in string.ascii_uppercase]
    + [
        (f"ctrl+{x}", ord(x) - (ord("a") - ord("A")), False, True)
        for x in string.ascii_lowercase
    ],
)
def test_fkey_letters(keyspec, key, is_shift, is_ctrl):

    keystroke = ic(FKey(keyspec))
    assert keystroke.key == key
    assert keystroke.is_shift is is_shift
    assert keystroke.is_ctrl is is_ctrl


def test_fkey_a_A():

    assert FKey("a").__dict__ == FKey("A").__dict__


def test_fkey_f1_F1():

    assert FKey("f1").__dict__ == FKey("F1").__dict__


def test_fkey_f1():

    keystroke = ic(FKey("f1"))
    assert not keystroke.is_shift
    assert not keystroke.is_ctrl


def test_fkey_f12():

    keystroke = ic(FKey("f12"))
    assert not keystroke.is_shift
    assert not keystroke.is_ctrl


def test_fkey_f13():

    keystroke = ic(FKey("f13"))
    assert not keystroke.is_shift
    assert not keystroke.is_ctrl


def test_fkey_shift_f1():

    keystroke = ic(FKey("shift+f1"))
    assert keystroke.is_shift
    assert not keystroke.is_ctrl


def test_fkey_shift_f12():

    keystroke = ic(FKey("shift+f12"))
    assert keystroke.is_shift
    assert not keystroke.is_ctrl


def test_fkey_shift_f13():

    keystroke = ic(FKey("shift+f13"))
    assert keystroke.is_shift
    assert not keystroke.is_ctrl


def test_fkey_ctrl_f1():

    keystroke = ic(FKey("ctrl+f1"))
    assert not keystroke.is_shift
    assert keystroke.is_ctrl


def test_fkey_ctrl_f12():

    keystroke = ic(FKey("ctrl+f12"))
    assert not keystroke.is_shift
    assert keystroke.is_ctrl


def test_fkey_ctrl_f13():

    keystroke = ic(FKey("ctrl+f13"))
    assert not keystroke.is_shift
    assert keystroke.is_ctrl



from pathlib import Path

import icecream
import pytest
from icecream import ic

from tf2mon.hacker import HackerManager

icecream.ic.configureOutput(prefix="=====>\n", includeContext=True)
CACHE = Path.home() / ".cache" / "tf2mon"
BASE = CACHE / "hackers-base.json"
LOCAL = CACHE / "hackers-local.json"
# pylint: disable=protected-access
# pylint: disable=using-constant-test

disabled = pytest.mark.skipif(True, reason="disabled")


# @disabled
def test_print_hacker_names():
    hackers = HackerManager(base=BASE, local=LOCAL)
    print(f"combined: there are {len(hackers._hackers_by_name):3} names")
    for name, list_of_hackers in hackers._hackers_by_name.items():
        print(f"combined: len={len(list_of_hackers):3} name=`{name}`")
        # for hacker in list_of_hackers:
        #    ic(hacker)


@disabled
def test_print_base_hacker_names():
    hackers = HackerManager(base=BASE)
    print(f"base: there are {len(hackers._hackers_by_name):3} names")
    for name, list_of_hackers in hackers._hackers_by_name.items():
        print(f"base: len={len(list_of_hackers):3} name=`{name}`")


@disabled
def test_print_local_hacker_names():
    hackers = HackerManager(local=LOCAL)
    print(f"local: there are {len(hackers._hackers_by_name):3} names")
    for name, list_of_hackers in hackers._hackers_by_name.items():
        print(f"local: len={len(list_of_hackers):3} name=`{name}`")


@disabled
def test_hacker():

    hackers = HackerManager(base=None, local=LOCAL)

    if False:
        for k, v in hackers._hackers_by_name.items():
            ic(k)
            for i in v:
                ic(i)

    if False:
        for k, v in hackers._hackers_by_steamid.items():
            print(f"k={k} v={v}")

    if False:
        for hacker in hackers._hackers_by_name.values():
            ic(hacker)

    if False:
        for hacker in hackers._hackers_by_steamid.values():
            ic(hacker)



"""Command.

May be invoked by:
    - keypress within game.
    - keypress within monitor.
    - commands typed into game console (`~`).
    - commands typed into monitor admin repl.
"""

import re
import time
from dataclasses import dataclass, field
from typing import Callable

from libcurses import register_fkey

from tf2mon import APPNAME, APPTAG
from tf2mon.fkey import FKey
from tf2mon.regex import Regex


@dataclass
class Command:
    """Command.

    Attributes:
        name: str
            Command name; e.g., 'HELP', 'TOGGLE-SORT'.

        status: Callable[..., str]
            must return current value as `str` for display;
            Default `None` to hide.

        handler: Callable[re.Match, None]
            must perform monitor function; Default `None` for no-op. e.g.,
            cmd="PULL", only has a game function to perform (which is to
            push `status`... it's pull from the monitor's perspective.)

        action: str
            game action to `exec`; e.g., cmd="KICKS-POP" has
            action="tf2mon_kicks_pop". Default `None` causes game to send
            event notice to monitor. e.g., cmd='TOGGLE-SORT' only has a
            monitor function to perform.

    """

    name: str
    status: Callable[..., str] = None
    handler: Callable[re.Match, None] = None
    action: str = None

    def __post_init__(self):
        """Init."""

        if not self.action:
            # Default, have game send this event notification message to
            # monitor whenever game calls for this command, such as in
            # response to an in-game key-press or mouse-click.
            self.action = f"echo {APPTAG}{self.name}"

    @property
    def request_pattern(self) -> str:
        """Return pattern to detect request from game to invoke this command."""

        # anchor head, optional timestamp, token, anchor tail.
        return r"^(\d{2}/\d{2}/\d{4} - \d{2}:\d{2}:\d{2}: )?" f"{APPTAG}{self.name}$"


@dataclass
class Function:
    """An `FKey`/`Command` binding."""

    keystroke: FKey
    command: Command

    def __repr__(self):
        return str(self.__dict__)


@dataclass
class Key:
    """A physical key may perform `base`, `ctrl` and `shift` `Function`s."""

    name: str  # e.g., "A", "F1"
    base: Function = field(default=None, init=False)
    ctrl: Function = field(default=None, init=False)
    shift: Function = field(default=None, init=False)

    @property
    def functions(self) -> [Function]:
        """Return list of `Function`s bound to this `Key`."""
        return [x for x in [self.base, self.ctrl, self.shift] if x is not None]


class CommandManager:
    """Command manager."""

    def __init__(self):
        """Init."""

        self.key_by_name: dict[str, Key] = {}

    def bind(
        self,
        cmd_factory: Callable[..., Command],
        keyspec: str,
        game_only: bool = False,
    ) -> None:
        """Bind `Command` returned by `cmd_factory` with `keyspec`."""

        command = cmd_factory()
        if game_only:
            command.handler = None

        keystroke = FKey(keyspec)
        if (key := self.key_by_name.get(keystroke.name)) is None:
            key = Key(keystroke.name)

        function = Function(keystroke, command)

        if keystroke.is_ctrl:
            if key.ctrl:
                raise ValueError("duplicate keyspec", keyspec)
            key.ctrl = function
        #
        elif keystroke.is_shift:
            if key.shift:
                raise ValueError("duplicate keyspec", keyspec)
            key.shift = function
        #
        else:
            if key.base:
                raise ValueError("duplicate keyspec", keyspec)
            key.base = function

        self.key_by_name[keystroke.name] = key

    def as_tf2_exec_script(self, static: str, dynamic: str) -> str:
        """Return game key bindings as TF2 exec script."""

        lines = [
            f"// Auto-generated by {APPNAME} on {time.asctime()}",
            "",
            'echo "-------------------------------------------------------------------------"',
            f'echo "Running {static}"',
            "",
            "// Push from game to monitor.",
            'alias tf2mon_push               "status; tf_lobby_debug"',
            "",
            "// Pull from monitor into game.",
            f'alias tf2mon_pull               "exec {dynamic}"',
            "",
            "// which constantly redefines these aliases:",
            'alias tf2mon_kicks_pop          "tf2mon_pull; _tf2mon_kicks_pop"',
            'alias tf2mon_kicks_popleft      "tf2mon_pull; _tf2mon_kicks_popleft"',
            'alias tf2mon_spams_pop          "tf2mon_pull; _tf2mon_spams_pop"',
            'alias tf2mon_spams_popleft      "tf2mon_pull; _tf2mon_spams_popleft"',
            "",
            f'alias tf2mon_kicks_clear        "echo {APPTAG}KICKS-CLEAR"',
            f'alias tf2mon_spams_clear        "echo {APPTAG}SPAMS-CLEAR"',
            f'alias tf2mon_clear_queues       "echo {APPTAG}CLEAR-QUEUES"',
            "",
            "// Bind unmodified keys directly to their `base` commands.",
        ]

        for key in [
            x for x in self.key_by_name.values() if x.name and x.base and not (x.ctrl or x.shift)
        ]:
            if key.base.command.action:
                lines.append(f'bind "{key.name}" "{key.base.command.action}"')

        lines.extend(
            [
                "",
                "// Bind modified keys indirectly through aliases.",
            ]
        )

        _base = ["_user_bind_base", "_class_bind_base"]
        _ctrl = ["_user_bind_ctrl", "_class_bind_ctrl"]
        _shift = ["_user_bind_shift", "_class_bind_shift"]

        for key in [x for x in self.key_by_name.values() if x.name and (x.ctrl or x.shift)]:
            if key.base and key.base.command.action:
                alias = f"_b{key.name}"
                lines.append(f'alias {alias} "{key.base.command.action}"')
                _base.append(f"bind {key.name} {alias}")

            if key.ctrl and key.ctrl.command.action:
                alias = f"_c{key.name}"
                lines.append(f'alias {alias} "{key.ctrl.command.action}"')
                _ctrl.append(f"bind {key.name} {alias}")

            if key.shift and key.shift.command.action:
                alias = f"_s{key.name}"
                lines.append(f'alias {alias} "{key.shift.command.action}"')
                _shift.append(f"bind {key.name} {alias}")

        lines.extend(
            [
                "",
                "// User class scripts may hook in through these;",
                "// e.g., `.tf2/cfg/user/engineer.cfg` could:",
                '//   alias _class_bind_base "bind 1 slot1; bind 2 slot2" etc.',
                '//   alias _class_bind_shift "bind 1 build_sentry; bind 2 build_dispenser" etc',
                'alias _class_bind_base ""',
                'alias _class_bind_ctrl ""',
                'alias _class_bind_shift ""',
                "",
                "// User scripts may hook in through these.",
                "// e.g., `.tf2/cfg/user/bad-dad-aliases.cfg` could:",
                '//   alias _user_bind_base "bind a +moveleft"',
                '//   alias _user_bind_shift "bind a load_itempreset 0"',
                'alias _user_bind_base ""',
                'alias _user_bind_ctrl ""',
                'alias _user_bind_shift ""',
                "",
                "// Toggle modified keys.",
            ]
        )

        # alias _bind_base "bind f1 _bf1; bind f2 _bf2; bind f3 _bf3; bind f4 _bf4"
        # alias _bind_ctrl "bind f1 _cf1; bind f2 _cf2; bind f3 _cf3; bind f4 _cf4"
        # alias _bind_shift "bind f1 _sf1; bind f2 _sf2; bind f3 _sf3; bind f4 _sf4"

        if binds := "; ".join(_base):
            lines.append(f'alias _bind_base "{binds}"')

        if binds := "; ".join(_ctrl):
            lines.append(f'alias _bind_ctrl "{binds}"')

        if binds := "; ".join(_shift):
            lines.append(f'alias _bind_shift "{binds}"')

        lines.extend(
            [
                "",
                'bind ctrl "+ctrled"',
                "alias +ctrled _bind_ctrl",
                "alias -ctrled _base_base",
                "",
                'bind shift "+shifted"',
                "alias +shifted _bind_shift",
                "alias -shifted _bind_base",
                "",
                "_bind_base",
            ]
        )

        return "\n".join(lines)

    def register_curses_handlers(self):
        """Register curses key handlers."""

        for key in self.key_by_name.values():
            for function in key.functions:
                if function.command.handler:
                    register_fkey(function.command.handler, function.keystroke.key)

    def get_regex_list(self):
        """Return list of `Regex` for all handlers."""

        items = []
        for key in self.key_by_name.values():
            for function in key.functions:
                if function.command.handler:
                    items.append(
                        Regex(function.command.request_pattern, function.command.handler)
                    )
        return items

    def get_status_line(self):
        """Return 1-line string showing current state of function keys."""

        items = []
        for key in self.key_by_name.values():
            for function in key.functions:
                if function.command.status:
                    items.append(f"{function.keystroke.label}={function.command.status()}")
        return " ".join(items)



"""TF2's console logfile."""

import re
import time
from collections import namedtuple
from pathlib import Path

from loguru import logger

import tf2mon


class Conlog:
    """TF2 writes console output to the file named in its `con_logfile` variable.

    Options from our command line, and commands from our admin console, may
    "inject" lines into the data stream read from the game console logfile.
    """

    # pylint: disable=too-many-instance-attributes

    _CMD = namedtuple("_cmd", ["lineno", "cmd"])

    def __init__(
        self,
        path: Path | str,
        exclude_file: Path | str = None,
        *,
        rewind: bool = True,
        follow: bool = False,
        inject_cmds=None,
        inject_file=None,
    ):
        """Prepare to open and read the console logfile."""

        self.path = path
        self.rewind = rewind
        self.follow = follow

        self.is_eof: bool = False
        self.last_line = None
        self.lineno: int = 0

        logger.info(f"Reading `{exclude_file}`")
        self.re_exclude = re.compile(
            "|".join(exclude_file.read_text(encoding="utf-8").splitlines())
        )

        self._buffer: str = None
        self._file = None
        self._inject_cmds = []  # list(_CMD)
        self._is_inject_paused = False
        self._is_inject_sorted = False

        if inject_cmds:
            self._inject_cmd_list(inject_cmds)

        if inject_file:
            logger.info(f"Reading `{inject_file}`")
            with open(inject_file, encoding="utf-8") as file:
                self._inject_cmd_list(file)

    def _inject_cmd_list(self, cmds):
        """Inject list of commands into logfile."""

        for line in cmds:
            lineno, cmd = line.strip().split(":", maxsplit=1)
            self.inject_cmd(lineno, cmd)

    def inject_cmd(self, lineno, cmd):
        """Inject command into logfile before `lineno`."""

        if not cmd.startswith(tf2mon.APPTAG):
            cmd = tf2mon.APPTAG + cmd

        self._inject_cmds.append(self._CMD(int(lineno) - 1, cmd))
        self._is_inject_sorted = False

    def open(self):
        """Wait for existence of and open console logfile."""

        while not self.path.exists():
            logger.warning(f"Waiting for {str(self.path)!r}...")
            time.sleep(3)

        logger.info(f"Reading `{self.path}`")

        self._file = open(  # pylint: disable=consider-using-with
            self.path, encoding="utf-8", errors="replace"
        )

        if not self.rewind:
            while self._file.readline() != "":
                self.lineno += 1

            logger.log("ADMIN", f"lineno={self.lineno}")
            self.is_eof = True

    def readline(self):
        """Read and return next line from console lofgile.

        Return None on end-of-file, else line.strip() (which may evaluate False).
        """

        while True:

            if _buffer := self._buffer:
                self._buffer = None
                self.last_line = "{self.lineno}: {_buffer}"
                return _buffer

            if not self._is_inject_sorted:
                self._inject_cmds.sort(key=lambda x: x.lineno)
                self._is_inject_sorted = True

            if (
                self._inject_cmds
                and (self.is_eof or not self._is_inject_paused)
                and self._inject_cmds[0].lineno <= self.lineno
            ):

                line = self._inject_cmds.pop(0).cmd
                self._is_inject_paused = True
                self.last_line = "{self.lineno + 1}: {line}"
                logger.log("injected", self.last_line)
                return line

            self._is_inject_paused = False

            try:
                line = self._file.readline()
            except UnicodeDecodeError as err:
                logger.critical(err)
                continue

            if line:
                self.lineno += 1

                line = line.strip()
                if line.startswith(tf2mon.APPTAG) and " " in line:
                    # sometimes newlines get dropped and lines are combined
                    cmd, self._buffer = line.split(sep=" ", maxsplit=1)
                    self.last_line = f"{self.lineno}: {cmd}"
                    return cmd

                if self.re_exclude.search(line):
                    logger.log("exclude", f"{self.lineno}: {line}")
                    continue

                self.last_line = f"{self.lineno}: {line}"
                return line

            self.is_eof = True
            if not self.follow:
                return None  # eof

            time.sleep(1)  # hello

    def trunc(self):
        """Truncate console logfile."""

        with open(self.path, "w", encoding="utf-8"):
            pass

    def clean(self):
        """Print non-excluded lines to `stdout`."""

        with open(self.path, encoding="utf-8", errors="replace") as file:
            for line in [x for x in file if not self.re_exclude.search(x)]:
                print(line.rstrip("\n"))

    def _exclude_lines(self) -> [str]:
        """Return list of lines to exclude."""

        return [
            "Failed to find attachment point specified for",
            "# userid name",
            "##### CTexture::LoadTextureBitsFromFile",
            "CMaterialVar::GetVecValue: trying to get a vec value for",
            "Cannot update control point",
            "DataTable warning: tf_objective_resource: Out-of-range value",
            "EmitSound: pitch out of bounds",
            'Error: Material "models/workshop/player/items/all_class',
            "Lobby updated",
            "Missing Vgui material",
            'No such variable "',
            "Requesting texture value from var",
            "SOLID_VPHYSICS static prop with no vphysics model!",
            "SetupBones: invalid bone array size",
            "ertexlit_and_unlit_generic_bump_ps20b.vcs",
            "m_face->glyph->bitmap.width is 0 for ch:32",
            "to attach particle system",
            "to get specular",
            "unknown particle system",
        ]



"""Base class for display layout."""

import curses
from dataclasses import KW_ONLY, InitVar, dataclass

import libcurses


@dataclass(slots=True)
class BaseLayout:
    """Base class for display layout.

    Each layout must initialize:
        - logger_win
        - status_win
        - cmdline_win
    others may remain undefined.
    """

    # pylint: disable=too-many-instance-attributes

    grid: InitVar[libcurses.Grid]
    max_users: InitVar[int] = 32
    _: KW_ONLY
    chatwin_blu: curses.window = None
    chatwin_red: curses.window = None
    scorewin_blu: curses.window = None
    scorewin_red: curses.window = None
    user_win: curses.window = None
    kicks_win: curses.window = None
    spams_win: curses.window = None
    duels_win: curses.window = None
    logger_win: curses.window = None
    status_win: curses.window = None
    cmdline_win: curses.window = None

    def __post_init__(self, grid: libcurses.Grid, max_users: int):
        """Build windows."""
        _ = grid  # unused
        _ = max_users  # unused



"""Gameplay patterns/handlers."""

from loguru import logger

from tf2mon.chat import Chat
from tf2mon.hacker import HackerAttr
from tf2mon.regex import Regex
from tf2mon.steamplayer import steamid_from_str
from tf2mon.user import Team, UserState


class Gameplay:
    """Gameplay."""

    # pylint: disable=line-too-long

    def __init__(self, monitor):
        """Initialize Gameplay."""

        leader = (
            r"^(\d{2}/\d{2}/\d{4} - \d{2}:\d{2}:\d{2}: )?"  # anchor to head; optional timestamp
        )

        self.monitor = monitor

        self.regex_list = [
            # key_listboundkeys
            # Regex(
            #    '^".*" = ".*"$',
            #    lambda m: logger.debug('key_listboundkeys')),
            # new server
            Regex(
                leader + "(Client reached server_spawn.$|Connected to [0-9])",
                lambda m: self.monitor.reset_game(),
            ),
            # capture/defend
            Regex(
                leader
                + r"(?P<username>.*) (?P<action>(?:captured|defended)) (?P<capture_pt>.*) for team #(?P<teamno>\d)$",  # noqa
                lambda m: self.capture(
                    m.group("username"),
                    m.group("action"),
                    m.group("capture_pt"),
                    int(m.group("teamno")),
                ),
            ),
            # must be before `chat`
            # account : not logged in  (No account specified)
            # version : 6173888/24 6173888 secure
            # map     : pl_barnblitz at: 0 x, 0 y, 0 z
            # udp/ip  : 208.78.164.167:27067  (public ip: 208.78.164.167)
            # tags    : hidden,increased_maxplayers,payload,valve
            # steamid : [A:1:3814649857:15826] (90139968514486273)
            # players : 20 humans, 0 bots (32 max)
            # edicts  : 1378 used of 2048 max
            Regex(
                leader + r"(account|version|map|udp\/ip|tags|steamid|players|edicts)\s+: (.*)",
                lambda m: logger.log("server", m.group(0)),
            ),
            # "06/05/2022 - 13:54:19: Client ping times:"
            Regex(
                leader + r"Client ping times:",
                lambda m: logger.log("server", m.group(0)),
            ),
            # "06/05/2022 - 13:54:19:   67 ms : luft"
            # "06/05/2022 - 13:54:19:xy 87 ms : BananaHatTaco"
            Regex(
                leader + r"\s*\d+ ms .*",
                lambda m: logger.log("server", m.group(0)),
            ),
            # chat
            # 'Bad Dad :  hello'
            # '*DEAD* Bad Dad :  hello'
            # '*DEAD*(TEAM) Bad Dad :  hello'
            Regex(
                leader
                + r"(?:(?P<dead>\*DEAD\*)?(?P<teamflag>\(TEAM\))? )?(?P<username>.*) :  ?(?P<msg>.*)$",  # noqa
                lambda m: self.playerchat(
                    m.group("teamflag"), m.group("username"), m.group("msg")
                ),
            ),
            # kill
            Regex(
                leader
                + r"(?P<killer>.*) killed (?P<victim>.*) with (?P<weapon>.*)\.(?P<crit> \(crit\))?$",  # noqa
                lambda m: self.kill(
                    m.group("killer"), m.group("victim"), m.group("weapon"), m.group("crit")
                ),
            ),
            # connected
            Regex(
                leader + "(?P<username>.*) connected$",
                lambda m: self.connected(m.group("username")),
            ),
            # status
            # "# userid name                uniqueid            connected ping loss state"
            # "#     29 "Bad Dad"           [U:1:42708103]      01:24       67    0 active"
            Regex(
                leader
                + r'#\s*(?P<userid>\d+) "(?P<username>.+)"\s+(?P<steamid>\S+)\s+(?P<elapsed>[\d:]+)\s+(?P<ping>\d+)',  # noqa
                lambda m: self.status(
                    int(m.group("userid")),
                    m.group("username"),
                    m.group("steamid"),
                    m.group("ping"),
                ),
            ),
            # status
            # "# userid name                uniqueid            connected ping loss state"
            # "#      3 "Nobody"            BOT                                     active
            Regex(
                leader + r'#\s*(?P<userid>\d+) "(?P<username>.+)"\s+(?P<steamid>BOT)\s+active',
                lambda m: self.status(
                    int(m.group("userid")),
                    m.group("username"),
                    m.group("steamid"),
                    0,
                ),
            ),
            # tf_lobby_debug
            # "Member[22] [U:1:42708103]  team = TF_GC_TEAM_INVADERS  type = MATCH_PLAYER"
            Regex(
                leader
                + r"\s*(Member|Pending)\[\d+\] (?P<steamid>\S+)\s+team = (?P<teamname>\w+)",
                lambda m: self.lobby(m.group("steamid"), m.group("teamname")),
            ),
            Regex(
                leader + "Failed to find lobby shared object",
                lambda m: logger.trace("tf_lobby_debug failed: " + m.group(0)),
            ),
            #
            # Regex(
            #    '^Teams have been switched',
            #    lambda m: self.monitor.switch_teams()),
            #
            Regex(
                leader + r"You have switched to team (?P<teamname>\w+) and will",
                lambda m: self.monitor.me.assign_team(m.group("teamname")),
            ),
            # hostname: Valve Matchmaking Server (Virginia iad-1/srcds148 #53)
            Regex(
                "^hostname: (.*)",
                lambda m: (
                    logger.log("server", m.group(0)),
                    self.monitor.users.check_status(),
                ),
            ),
            # "FFD700[RTD] FF4040your mother rolled 32CD32PowerPlay."
            Regex(
                r"[0-9A-F]{6}\[RTD\] [0-9A-F]{6}(?P<username>.*) rolled [0-9A-F]{6}(?P<perk>.*)",  # noqa
                lambda m: self.perk(m.group("username"), m.group("perk")),
            ),
            Regex(
                r"[0-9A-F]{6}\[RTD\] [0-9A-F]{6}(?P<username>.*)\'s perk has worn off.",
                lambda m: self.perk(m.group("username"), None),
            ),
            Regex(
                r"[0-9A-F]{6}\[RTD\] [0-9A-F]{6}(?P<username>.*) has changed class during their roll.",  # noqa
                lambda m: self.perk(m.group("username"), None),
            ),
            Regex(
                r"[0-9A-F]{6}\[RTD\] Your perk has worn off.", lambda m: self.perk(None, None)
            ),
        ]

    def capture(self, username, action, capture_pt, teamno):
        """Handle message."""

        for name in username.split(", "):  # fix: names containing commas

            user = self.monitor.users.find_username(name)

            user.assign_teamno(teamno)

            if action == "captured":
                user.ncaptures += 1
                level = "CAP"
            else:
                user.ndefenses += 1
                level = "DEF"
            level += user.team.name

            logger.log(level, f"{user} {capture_pt!r}")
            user.actions.append(f"{level} {capture_pt!r}")

    def playerchat(self, teamflag, username, msg):
        """Handle message."""

        user = self.monitor.users.find_username(username)
        chat = Chat(user, teamflag, msg)

        user.chats.append(chat)
        self.monitor.chats.append(chat)
        self.monitor.ui.show_chat(chat)

        # if this is a team chat, then we know we're on the same team, and
        # if one of us knows which team we're on and the other doesn't, we
        # can assign.

        me = my = self.monitor.me
        if chat.teamflag and user != me:
            # we're on the same team
            if not user.team:
                if my.team:
                    user.assign_team(my.team)
            elif not my.team:
                me.assign_team(user.team)

        # inspect msg
        if self.monitor.is_racist_text(chat.msg):
            user.kick(HackerAttr.RACIST)

        # If this looks like a milenko chat, mark him to be tracked when
        # his steamid becomes available. Doing this now to notify the
        # operator asap to `TF2MON-PUSH` steamids to us.
        #
        # Why track them? No requirement; curious to see relationship
        # between names and steamids... one-to-one, many-to-one?

        elif user.is_milenko_chat(chat):
            user.kick(HackerAttr.MILENKO)

        elif user.is_cheater_chat(chat):
            user.kick(HackerAttr.CHEATER)

    def kill(self, s_killer: str, s_victim: str, weapon: str, s_crit: str) -> None:
        """Handle message."""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        killer = self.monitor.users.find_username(s_killer)
        victim = self.monitor.users.find_username(s_victim)

        killer.last_victim = victim
        victim.last_killer = killer

        # do most calculations now (once);
        # to avoid calculating when rendering scoreboard (often).

        killer.opponents[victim.key] = victim
        victim.opponents[killer.key] = killer

        killer.victims[victim.key] = victim
        victim.killers[killer.key] = killer

        # totals ---------------------------------------------------------------

        killer.nkills += 1
        victim.ndeaths += 1

        _k = killer.nkills
        _d = killer.ndeaths
        killer.kdratio = float(_k) if not _d else _k / _d

        _k = victim.nkills
        _d = victim.ndeaths
        victim.kdratio = float(_k) if not _d else _k / _d

        # subtotals by opponent-------------------------------------------------

        if victim.key not in killer.nkills_by_opponent:
            killer.nkills_by_opponent[victim.key] = 0
        killer.nkills_by_opponent[victim.key] += 1

        if killer.key not in victim.ndeaths_by_opponent:
            victim.ndeaths_by_opponent[killer.key] = 0
        victim.ndeaths_by_opponent[killer.key] += 1

        _k = killer.nkills_by_opponent.get(victim.key, 0)
        _d = killer.ndeaths_by_opponent.get(victim.key, 0)
        killer.kdratio_by_opponent[victim.key] = float(_k) if not _d else _k / _d

        _k = victim.nkills_by_opponent.get(killer.key, 0)
        _d = victim.ndeaths_by_opponent.get(killer.key, 0)
        victim.kdratio_by_opponent[killer.key] = float(_k) if not _d else _k / _d

        # subtotal opponents by weapon_state -----------------------------------

        if role := self.monitor.role_by_weapon.get(weapon):
            killer.role = role
            if killer.role == self.monitor.sniper_role:
                killer.nsnipes += 1
        else:
            role = killer.role
            if weapon not in ("player", "world"):
                logger.error(f"cannot map {weapon} for {killer}")

        crit = bool(s_crit)
        weapon_state = role.get_weapon_state(weapon, crit, killer.perk)

        if victim.key not in killer.nkills_by_opponent_by_weapon:
            # contains a hash of counts by weapon_state
            killer.nkills_by_opponent_by_weapon[victim.key] = {}
        if weapon_state not in killer.nkills_by_opponent_by_weapon[victim.key]:
            killer.nkills_by_opponent_by_weapon[victim.key][weapon_state] = 0
        killer.nkills_by_opponent_by_weapon[victim.key][weapon_state] += 1

        #
        level = "KILL"
        if killer.team:
            level += killer.team.name

        logger.log(
            level,
            "killer {!r} victim {!r} weapon {!r}",
            killer.moniker,
            victim.moniker,
            weapon_state,
        )

        if killer == self.monitor.me:
            self.monitor.spammer.taunt(victim, weapon, crit)

        if victim == self.monitor.me:
            self.monitor.spammer.throe(killer, weapon, crit)

        if not victim.team and killer.team:
            victim.assign_team(killer.opposing_team)
        elif not killer.team and victim.team:
            killer.assign_team(victim.opposing_team)

    def connected(self, username):
        """Handle message."""

        logger.log("CONNECT", self.monitor.users.find_username(username))

        self.monitor.ui.notify_operator = True

    def status(self, userid, username, s_steamid, ping):
        """Handle message."""

        self.monitor.ui.notify_operator = False
        if not (steamid := steamid_from_str(s_steamid)):
            return  # invalid

        self.monitor.users.status(userid, username, steamid, ping)

    def lobby(self, s_steamid, teamname):
        """Handle message."""

        # this will not be called for games on local server with bots
        # or community servers; only on valve matchmaking servers.

        if not (steamid := steamid_from_str(s_steamid)):
            return  # invalid

        if teamname == "TF_GC_TEAM_INVADERS":
            team = Team.BLU
        elif teamname == "TF_GC_TEAM_DEFENDERS":
            team = Team.RED
        else:
            logger.critical(f"bad teamname {teamname!r} steamid {steamid}")
            return

        self.monitor.users.lobby(steamid, team)

    def perk(self, username, perk):
        """Handle message."""

        user = self.monitor.users.find_username(username) if username else self.monitor.me

        if perk:
            logger.log("PERK-ON", f"{user} {perk!r}")
        else:
            logger.log("PERK-OFF", f"{user} {user.perk!r}")

        user.perk = perk

    def repl(self):
        """Read the console log file and play game."""

        self.monitor.steam_web_api.connect()
        self.monitor.conlog.open()

        #
        while (line := self.monitor.conlog.readline()) is not None:

            if not line:  # blank line
                continue

            regex = Regex.search_list(line, self.monitor.regex_list)
            if not regex:
                logger.log("ignore", self.monitor.conlog.last_line)
                continue

            self.monitor.admin.step(line)

            regex.handler(regex.re_match_obj)

            # Vet all unvetted users that can be vetted, and perform all
            # postponed work that can be performed.

            for user in self.monitor.users.active_users():

                if not user.vetted and user.steamid:
                    user.vet_player()

                if user.vetted and user.work_attr:
                    user.kick()

                if user.state == UserState.DELETE:
                    self.monitor.users.delete(user)

            # push work to the game
            self.monitor.msgqueues.send()

            #
            self.monitor.ui.update_display()
